"""
MockAPI, a fully featured API/host routing framework for mitmproxy,
ease your life from if-else.

"""

import os
import sys
import traceback
import urllib.parse
from typing import List, Optional, Tuple, Union
from parse import Parser
from mitmproxy.connection import Server
from mitmproxy.http import HTTPFlow, Response
from mitmproxy.net.http import url

import functools
import logging
import re

__author__ = "ttimasdf"
__copyright__ = "ttimasdf"
__license__ = "Apache-2.0"


class InterceptedAPI:
    REQUEST = 0x01
    RESPONSE = 0x02
    META_REQ_PASSTHROUGH = "mockapi-request-passthrough"
    META_RESP_PASSTHROUGH = "mockapi-response-passthrough"
    META_REQ_URLPARSE = "mockapi-request-urlparse"
    META_REQ_HOST = "mockapi-request-host"

    REGEX_HOST_HEADER = re.compile(r"^(?P<host>[^:]+|\[.+\])(?::(?P<port>\d+))?$")

    PROXY_FORWARDED_HEADERS = [
        "X-Forwarded-For",
        "X-Forwarded-Host",
        "X-Forwarded-Port",
        "X-Forwarded-Proto",
        "X-Forwarded-Server",
        "X-Real-Ip",
    ]

    def __init__(self,
            default_host: Optional[str]=None,
            host_mapping: List[Tuple[Union[str, re.Pattern], str]]={},
            blacklist_domain:List[str]=[],
            request_passthrough: bool=True,
            response_passthrough: bool=True,
            respect_proxy_headers: bool=False):

        self.default_host = default_host
        self.host_mapping = host_mapping
        self.request_routes: List[Tuple[Optional[str], Parser, callable]] = []
        self.response_routes: List[Tuple[Optional[str], Parser, callable]] = []
        self.blacklist_domain = blacklist_domain
        self.request_passthrough = request_passthrough
        self.response_passthrough = response_passthrough
        self.respect_proxy_headers = respect_proxy_headers

        self._log = logging.getLogger("MockAPI")
        if os.getenv("MOCKAPI_LOG_DEBUG"):
            self._log.setLevel(logging.DEBUG)
        self._log.info("%s started", self.__class__.__name__)

    # def server_connect(self, data: ServerConnectionHookData):
    #     self._log.debug("Getting connection: peer=%s sock=%s addr=%s . state=%s",
    #         data.server.peername, data.server.sockname, data.server.address, data.server)

    def request(self, flow: HTTPFlow):
        if InterceptedAPI.META_REQ_URLPARSE in flow.metadata:
            url = flow.metadata[InterceptedAPI.META_REQ_URLPARSE]
        else:
            url = urllib.parse.urlparse(flow.request.path)
            flow.metadata[InterceptedAPI.META_REQ_URLPARSE] = url
        path = url.path
        if flow.metadata.get(InterceptedAPI.META_REQ_PASSTHROUGH) is True:
            self._log.warning("<= [%s] %s skipped because of previous passthrough", flow.request.method, path)
            return
        host = self.remap_host(flow)
        handler, params = self.find_handler(host, path, InterceptedAPI.REQUEST)
        if handler is not None:
            self._log.info("<= [%s] %s", flow.request.method, path)
            handler(flow, *params.fixed, **params.named)
        elif not self.request_passthrough or self.get_host(flow)[0] in self.blacklist_domain:
            self._log.warning("<= [%s] %s default response", flow.request.method, path)
            flow.response = self.default_response()
        else:
            flow.metadata[InterceptedAPI.META_REQ_PASSTHROUGH] = True
            self._log.debug("<= [%s] %s passthrough", flow.request.method, path)

    def response(self, flow: HTTPFlow):
        if InterceptedAPI.META_REQ_URLPARSE in flow.metadata:
            url = flow.metadata[InterceptedAPI.META_REQ_URLPARSE]
        else:
            url = urllib.parse.urlparse(flow.request.path)
            flow.metadata[InterceptedAPI.META_REQ_URLPARSE] = url
        path = url.path
        if flow.metadata.get(InterceptedAPI.META_RESP_PASSTHROUGH) is True:
            self._log.warning("=> [%s] %s skipped because of previous passthrough", flow.response.status_code, path)
            return
        handler, params = self.find_handler(self.get_host(flow)[0], path, InterceptedAPI.RESPONSE)
        if handler is not None:
            self._log.info("=> [%s] %s", flow.response.status_code, path)
            handler(flow, *params.fixed, **params.named)
        elif not self.response_passthrough or self.get_host(flow)[0] in self.blacklist_domain:
            self._log.warning("=> [%s] %s default response", flow.response.status_code, path)
            flow.response = self.default_response()
        else:
            flow.metadata[InterceptedAPI.META_RESP_PASSTHROUGH] = True
            self._log.debug("=> [%s] %s passthrough", flow.response.status_code, path)

    def route(self, path, host=None, reqtype=REQUEST, catch_error=True, return_error=False):
        host = host or self.default_host

        # generic catch-all wrapper
        def catcher(func):
            @functools.wraps(func)
            def handler(flow: HTTPFlow, *args, **kwargs):
                try:
                    return func(flow, *args, **kwargs)
                except Exception as e:
                    etype, value, tback = sys.exc_info()
                    tb = "".join(traceback.format_exception(etype, value, tback))
                    self._log.error(
                        "Exception catched when handling response to %s:\n%s",
                        flow.request.pretty_url,
                        tb,
                    )
                    if return_error:
                        flow.response = self.error_response(str(e))

            return handler

        def wrapper(handler):
            if catch_error:
                handler = catcher(handler)
            if reqtype == InterceptedAPI.REQUEST:
                self.request_routes.append((host, Parser(path), handler))
            elif reqtype == InterceptedAPI.RESPONSE:
                self.response_routes.append((host, Parser(path), handler))
            else:
                raise ValueError(f"Invalid route type: {reqtype}")
            return handler

        return wrapper

    def remap_host(self, flow: HTTPFlow, overwrite=True):
        host, port = self.get_host(flow)
        for src, dest in self.host_mapping:
            if (isinstance(src, re.Pattern) and src.match(host)) or \
                (isinstance(src, str) and host == src):
                if overwrite and (flow.request.host != dest or flow.request.port != port):
                    if self.respect_proxy_headers:
                        flow.request.scheme = flow.request.headers["X-Forwarded-Proto"]
                    flow.server_conn = Server((dest, port))
                    flow.request.host = dest
                    flow.request.port = port
                self._log.debug("flow: %s, remapping host: %s -> %s, port: %d", flow, host, dest, port)
                return dest
        return host

    def get_host(self, flow: HTTPFlow) -> Tuple[str, int]:
        if InterceptedAPI.META_REQ_HOST not in flow.metadata:
            if self.respect_proxy_headers:
                # all(h in flow.request.headers for h in ["X-Forwarded-Host", "X-Forwarded-Port"])
                host = flow.request.headers["X-Forwarded-Host"]
                port = int(flow.request.headers["X-Forwarded-Port"])
            else:
                # Get Destination Host
                host, port = url.parse_authority(flow.request.pretty_host, check=False)
                port = port or url.default_port(flow.request.scheme) or 80
            flow.metadata[InterceptedAPI.META_REQ_HOST] = (host, port)
        return flow.metadata[InterceptedAPI.META_REQ_HOST]

    def default_response(self):
        return Response.make(404, "Not Found", {"X-Intercepted-By": "mitmproxy"})

    def error_response(self, msg="APIServer Error"):
        return Response.make(502, msg)

    def find_handler(self, host, path, reqtype=REQUEST):
        if reqtype == InterceptedAPI.REQUEST:
            routes = self.request_routes
        elif reqtype == InterceptedAPI.RESPONSE:
            routes = self.response_routes
        else:
            raise ValueError(f"Invalid route type: {reqtype}")

        for h, parser, handler in routes:
            if h != host:
                continue
            parse_result = parser.parse(path)
            self._log.debug("Parse %s => %s", path, parse_result)
            if parse_result is not None:
                return handler, parse_result

        return None, None



"""
    Dummy conftest.py for xepor.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    - https://docs.pytest.org/en/stable/fixture.html
    - https://docs.pytest.org/en/stable/writing_plugins.html

    Imported from:
    https://github1s.com/mitmproxy/mitmproxy/blob/v7.0.4/test/conftest.py
"""

import os
import socket

from mitmproxy.utils import data

import pytest

# pytest_plugins = ('test.full_coverage_plugin',)

skip_windows = pytest.mark.skipif(
    os.name == "nt",
    reason='Skipping due to Windows'
)

skip_not_windows = pytest.mark.skipif(
    os.name != "nt",
    reason='Skipping due to not Windows'
)

try:
    s = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
    s.bind(("::1", 0))
    s.close()
except OSError:
    no_ipv6 = True
else:
    no_ipv6 = False

skip_no_ipv6 = pytest.mark.skipif(
    no_ipv6,
    reason='Host has no IPv6 support'
)


@pytest.fixture()
def tdata():
    return data.Data(__name__)



"""
Imported from:
https://github1s.com/mitmproxy/mitmproxy/blob/v7.0.4/test/mitmproxy/tservers.py
"""
from unittest import mock

from mitmproxy import controller
from mitmproxy import eventsequence
from mitmproxy import io
from mitmproxy.proxy import server_hooks
from mitmproxy.test import tflow
from mitmproxy.test import tutils


class MasterTest:

    async def cycle(self, master, content):
        f = tflow.tflow(req=tutils.treq(content=content))
        layer = mock.Mock("mitmproxy.proxy.protocol.base.Layer")
        layer.client_conn = f.client_conn
        layer.reply = controller.DummyReply()
        await master.addons.handle_lifecycle(server_hooks.ClientConnectedHook(layer))
        for e in eventsequence.iterate(f):
            await master.addons.handle_lifecycle(e)
        await master.addons.handle_lifecycle(server_hooks.ClientDisconnectedHook(layer))
        return f

    async def dummy_cycle(self, master, n, content):
        for i in range(n):
            await self.cycle(master, content)
        await master._shutdown()

    def flowfile(self, path):
        with open(path, "wb") as f:
            fw = io.FlowWriter(f)
            t = tflow.tflow(resp=True)
            fw.add(t)



import re
import pytest
from mitmproxy.http import Response
from mitmproxy.test import taddons, tflow

from .tserver import MasterTest

__author__ = "ttimasdf"
__copyright__ = "ttimasdf"
__license__ = "Apache-2.0"


class TestScripts(MasterTest):
    def test_httpbin(self, tdata):
        with taddons.context() as tctx:
            s = tctx.script(tdata.path("../examples/httpbin/httpbin.py"))
            api = s.addons[0]

            flow = tflow.tflow()
            flow.request.url = "http://httpbin.org/get"
            assert flow.response is None

            api.request(flow)
            assert flow.request.query["payload"] == "evil_param"

    def test_krisp_forward(self, tdata):
        with taddons.context() as tctx:
            s = tctx.script(tdata.path("../examples/krisp/krisp.py"))
            api = s.addons[0]

            flow = tflow.tflow()
            flow.request.url = "https://api.krisp.ai/random-path"
            assert flow.response is None

            api.request(flow)
            assert re.match(
                r"\d+\.\d+\.\d+\.\d+", flow.request.headers["X-Forwarded-For"]
            )

    def test_krisp_analytics(self, tdata):
        with taddons.context() as tctx:
            s = tctx.script(tdata.path("../examples/krisp/krisp.py"))
            api = s.addons[0]

            flow = tflow.tflow()
            flow.request.url = "https://analytics.krisp.ai/random-path-as-well"
            flow.request.set_content(rb'{"krisp": "analytics"}')
            assert flow.response is None

            api.request(flow)
            assert flow.response.headers["Content-Type"] == "application/json"

            payload = flow.response.json()
            assert payload["message"] == "Success"
            assert payload["data"]["app"]["krisp"] == "analytics"

    @pytest.mark.parametrize(
        "url,resp,stdout",
        [
            ("http://example.com/test", b"", ""),  # Nonsense url, should return nothing
            (  # step1
                f"http://player.polyv.net/videojson/cwqzcxkvj0iukomqxu0l591u2dke4vkc_c.json",
                b'{"body": "deadbeef"}',
                "Decrypted videojson for Test Lesson Title",
            ),
            (  # step2
                f"http://hls.videocc.net/jkag324wd2/e/cwqzcxkvj0iukomqxu0l591u2dke4vkc_1.m3u8?pid=544927262345396034173&device=desktop",
                b"",
                f"Found playlist for 123456 Test Lesson Title",
            ),
            (  # step3
                f"http://hls.videocc.net/playsafe/jkag324wd2/e/cwqzcxkvj0iukomqxu0l591u2dke4vkc_1.key?pid=544927262345396034173&device=desktop",
                b"",
                f"Found key for 123456 Test Lesson Title",
            ),
        ],
    )
    def test_polyv(self, capsys, monkeypatch, tdata, url, resp, stdout):
        with taddons.context() as tctx:
            polyv = tctx.script(tdata.path("../examples/polyv_scrapper/polyv.py"))
            api = polyv.addons[0]

            # Add test data
            polyv.ctx.lessons["cwqzcxkvj0iukomqxu0l591u2dke4vkc"] = polyv.Lesson(
                "cwqzcxkvj0iukomqxu0l591u2dke4vkc",
                "Test Lesson Title",
                "test/category",
                123456,
                1,
            )

            def monkey_decrypt(*args):
                import json
                import base64

                payload = {polyv.REDACTED_CONSTANT: "whosyourdaddy"}
                return base64.b64encode(json.dumps(payload).encode())

            def monkey_process_lesson(lesson):
                print("process_lesson is running")

            monkeypatch.setattr(polyv, "decrypt", monkey_decrypt)
            monkeypatch.setattr(polyv, "process_lesson", monkey_process_lesson)

            flow = tflow.tflow()
            flow.request.url = url
            flow.response = Response.make(content=resp)

            api.response(flow)
            captured = capsys.readouterr()
            assert stdout in captured.out



import pytest
from mitmproxy.http import HTTPFlow, Response
from mitmproxy.test import taddons, tflow
from xepor import InterceptedAPI


__author__ = "ttimasdf"
__copyright__ = "ttimasdf"
__license__ = "Apache-2.0"


@pytest.fixture
def api_simple():
    api = InterceptedAPI("example.com")

    @api.route("/test")
    def route1(flow: HTTPFlow):
        flow.response = Response.make(200, "TEST intercepted")

    @api.route(
        "/{}/{}/{vid}_1.m3u8",
        "hls.videocc.net",
    )
    def route2(flow: HTTPFlow, *args, vid):
        flow.response = Response.make(200, "TEST 2 INTERCEPTED")

    return api


@pytest.mark.parametrize(
    "req_url,resp_body",
    [
        ("http://example.com/test", "TEST intercepted"),
        ("http://hls.videocc.net/jkag324wd2/e/cwqzcxkvj0iukomqxu0l591u2dke4vkc_1.m3u8", "TEST 2 INTERCEPTED"),
    ],
)
def test_intercepted_route(api_simple, req_url, resp_body):
    with taddons.context(api_simple) as tctx:
        flow = tflow.tflow()
        flow.request.url = req_url
        assert flow.response is None

        api_simple.request(flow)
        assert resp_body in flow.response.text


def test_non_intercepted_route(api_simple):
    with taddons.context(api_simple) as tctx:
        flow = tflow.tflow()
        flow.request.path = "/test2"
        assert flow.response is None

        api_simple.request(flow)
        assert flow.response is None
